from unittest import TestCase

from aiortc.codecs import get_decoder, get_encoder
from aiortc.codecs.vpx import (VpxDecoder, VpxEncoder, VpxPayloadDescriptor,
                               _vpx_assert, number_of_threads)
from aiortc.mediastreams import VideoFrame
from aiortc.rtcrtpparameters import RTCRtpCodecParameters

VP8_CODEC = RTCRtpCodecParameters(name='VP8', clockRate=90000)


class VpxPayloadDescriptorTest(TestCase):
    def test_no_picture_id(self):
        descr, rest = VpxPayloadDescriptor.parse(b'\x10')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, None)
        self.assertEqual(descr.tl0picidx, None)
        self.assertEqual(descr.tid, None)
        self.assertEqual(descr.keyidx, None)
        self.assertEqual(bytes(descr), b'\x10')
        self.assertEqual(repr(descr), 'VpxPayloadDescriptor(S=1, PID=0, pic_id=None)')

        self.assertEqual(rest, b'')

    def test_short_picture_id_17(self):
        """
        From RFC 7741 - 4.6.3
        """
        descr, rest = VpxPayloadDescriptor.parse(b'\x90\x80\x11')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, 17)
        self.assertEqual(descr.tl0picidx, None)
        self.assertEqual(descr.tid, None)
        self.assertEqual(descr.keyidx, None)
        self.assertEqual(bytes(descr), b'\x90\x80\x11')
        self.assertEqual(repr(descr), 'VpxPayloadDescriptor(S=1, PID=0, pic_id=17)')

        self.assertEqual(rest, b'')

    def test_short_picture_id_127(self):
        descr, rest = VpxPayloadDescriptor.parse(b'\x90\x80\x7f')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, 127)
        self.assertEqual(descr.tl0picidx, None)
        self.assertEqual(descr.tid, None)
        self.assertEqual(descr.keyidx, None)
        self.assertEqual(bytes(descr), b'\x90\x80\x7f')

        self.assertEqual(rest, b'')

    def test_long_picture_id_128(self):
        descr, rest = VpxPayloadDescriptor.parse(b'\x90\x80\x80\x80')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, 128)
        self.assertEqual(descr.tl0picidx, None)
        self.assertEqual(descr.tid, None)
        self.assertEqual(descr.keyidx, None)
        self.assertEqual(bytes(descr), b'\x90\x80\x80\x80')

        self.assertEqual(rest, b'')

    def test_long_picture_id_4711(self):
        """
        From RFC 7741 - 4.6.5
        """
        descr, rest = VpxPayloadDescriptor.parse(b'\x90\x80\x92\x67')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, 4711)
        self.assertEqual(descr.tl0picidx, None)
        self.assertEqual(descr.tid, None)
        self.assertEqual(descr.keyidx, None)
        self.assertEqual(bytes(descr), b'\x90\x80\x92\x67')

        self.assertEqual(rest, b'')

    def test_tl0picidx(self):
        descr, rest = VpxPayloadDescriptor.parse(b'\x90\xc0\x92\x67\x81')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, 4711)
        self.assertEqual(descr.tl0picidx, 129)
        self.assertEqual(descr.tid, None)
        self.assertEqual(descr.keyidx, None)
        self.assertEqual(bytes(descr), b'\x90\xc0\x92\x67\x81')

        self.assertEqual(rest, b'')

    def test_tid(self):
        descr, rest = VpxPayloadDescriptor.parse(b'\x90\x20\xe0')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, None)
        self.assertEqual(descr.tl0picidx, None)
        self.assertEqual(descr.tid, (3, 1))
        self.assertEqual(descr.keyidx, None)
        self.assertEqual(bytes(descr), b'\x90\x20\xe0')

        self.assertEqual(rest, b'')

    def test_keyidx(self):
        descr, rest = VpxPayloadDescriptor.parse(b'\x90\x10\x1f')
        self.assertEqual(descr.partition_start, 1)
        self.assertEqual(descr.partition_id, 0)
        self.assertEqual(descr.picture_id, None)
        self.assertEqual(descr.tl0picidx, None)
        self.assertEqual(descr.tid, None)
        self.assertEqual(descr.keyidx, 31)
        self.assertEqual(bytes(descr), b'\x90\x10\x1f')

        self.assertEqual(rest, b'')


class Vp8Test(TestCase):
    def test_assert(self):
        with self.assertRaises(Exception) as cm:
            _vpx_assert(1)
        self.assertEqual(str(cm.exception), 'libvpx error: Unspecified internal error')

    def test_decoder(self):
        decoder = get_decoder(VP8_CODEC)
        self.assertTrue(isinstance(decoder, VpxDecoder))

    def test_encoder(self):
        encoder = get_encoder(VP8_CODEC)
        self.assertTrue(isinstance(encoder, VpxEncoder))

        frame = VideoFrame(width=640, height=480)
        payloads = encoder.encode(frame)
        self.assertEqual(len(payloads), 1)
        self.assertTrue(len(payloads[0]) < 1300)

        frame = VideoFrame(width=320, height=240)
        payloads = encoder.encode(frame)
        self.assertEqual(len(payloads), 1)
        self.assertTrue(len(payloads[0]) < 1300)

    def test_encoder_large(self):
        encoder = get_encoder(VP8_CODEC)
        self.assertTrue(isinstance(encoder, VpxEncoder))

        frame = VideoFrame(width=2560, height=1920)
        payloads = encoder.encode(frame)
        self.assertEqual(len(payloads), 7)
        self.assertEqual(len(payloads[0]), 1300)

    def test_number_of_threads(self):
        self.assertEqual(number_of_threads(1920 * 1080, 16), 8)
        self.assertEqual(number_of_threads(1920 * 1080, 8), 3)
        self.assertEqual(number_of_threads(1920 * 1080, 4), 2)
        self.assertEqual(number_of_threads(1920 * 1080, 2), 1)



import argparse
import asyncio
import json
import logging

from aiortc import RTCPeerConnection, RTCSessionDescription


def channel_log(channel, t, message):
    print('channel(%s) %s %s' % (channel.label, t, message))


def channel_watch(channel):
    @channel.on('message')
    def on_message(message):
        channel_log(channel, '<', message)


def create_pc():
    pc = RTCPeerConnection()

    @pc.on('datachannel')
    def on_datachannel(channel):
        channel_log(channel, '-', 'created by remote party')
        channel_watch(channel)

    return pc


async def run_answer(pc):
    done = asyncio.Event()

    @pc.on('datachannel')
    def on_datachannel(channel):
        @channel.on('message')
        def on_message(message):
            # reply
            message = 'pong'
            channel_log(channel, '>', message)
            channel.send(message)

            # quit
            done.set()

    # receive offer
    print('-- Please enter remote offer --')
    offer_json = json.loads(input())
    await pc.setRemoteDescription(RTCSessionDescription(
        sdp=offer_json['sdp'],
        type=offer_json['type']))
    print()

    # send answer
    await pc.setLocalDescription(await pc.createAnswer())
    answer = pc.localDescription
    print('-- Your answer --')
    print(json.dumps({
        'sdp': answer.sdp,
        'type': answer.type
    }))
    print()

    await done.wait()


async def run_offer(pc):
    done = asyncio.Event()

    channel = pc.createDataChannel('chat')
    channel_log(channel, '-', 'created by local party')
    channel_watch(channel)

    @channel.on('message')
    def on_message(message):
        # quit
        done.set()

    # send offer
    await pc.setLocalDescription(await pc.createOffer())
    offer = pc.localDescription
    print('-- Your offer --')
    print(json.dumps({
        'sdp': offer.sdp,
        'type': offer.type
    }))
    print()

    # receive answer
    print('-- Please enter remote answer --')
    answer_json = json.loads(input())
    await pc.setRemoteDescription(RTCSessionDescription(
        sdp=answer_json['sdp'],
        type=answer_json['type']))
    print()

    # send message
    message = 'ping'
    channel_log(channel, '>', message)
    channel.send(message)

    await done.wait()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Data channels with copy-and-paste signaling')
    parser.add_argument('role', choices=['offer', 'answer'])
    parser.add_argument('--verbose', '-v', action='count')
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    pc = create_pc()
    if args.role == 'offer':
        coro = run_offer(pc)
    else:
        coro = run_answer(pc)

    # run event loop
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(coro)
    except KeyboardInterrupt:
        pass
    finally:
        loop.run_until_complete(pc.close())



import argparse
import asyncio
import json
import logging
import random

import aiohttp
import websockets

from aiortc import (AudioStreamTrack, RTCPeerConnection, RTCSessionDescription,
                    VideoStreamTrack)
from aiortc.sdp import candidate_from_sdp


def description_to_dict(description):
    return {
        'sdp': description.sdp,
        'type': description.type
    }


class Signaling:
    async def connect(self, params):
        self.websocket = await websockets.connect(params['wss_url'], extra_headers={
            'Origin': 'https://appr.tc'
        })

    async def recv(self):
        data = await self.websocket.recv()
        return json.loads(data)

    async def send(self, data):
        await self.websocket.send(json.dumps(data))

    async def send_message(self, message):
        print('>', message)
        await self.send({
            'cmd': 'send',
            'msg': json.dumps(message)
        })


async def consume_signaling(signaling, pc, params):
    async def handle_message(message):
        print('<', message)

        if message['type'] == 'bye':
            return True

        if message['type'] == 'offer':
            await pc.setRemoteDescription(RTCSessionDescription(**message))
            await pc.setLocalDescription(await pc.createAnswer())
            await signaling.send_message(description_to_dict(pc.localDescription))
        elif message['type'] == 'answer':
            await pc.setRemoteDescription(RTCSessionDescription(**message))
        elif message['type'] == 'candidate':
            candidate = candidate_from_sdp(message['candidate'].split(':', 1)[1])
            candidate.sdpMid = message['id']
            candidate.sdpMLineIndex = message['label']
            pc.addIceCandidate(candidate)
        return False

    for data in params['messages']:
        message = json.loads(data)
        await handle_message(message)

    stop = False
    while not stop:
        data = await signaling.recv()
        message = json.loads(data['msg'])
        stop = await handle_message(message)


async def consume_audio(track):
    """
    Drain incoming audio.
    """
    while True:
        await track.recv()


async def consume_video(track):
    """
    Drain incoming video.
    """
    while True:
        await track.recv()


async def join_room(room):
    consumers = []

    # fetch room parameters
    async with aiohttp.ClientSession() as session:
        async with session.post('https://appr.tc/join/' + room) as response:
            # we cannot use response.json() due to:
            # https://github.com/webrtc/apprtc/issues/562
            data = json.loads(await response.text())
    assert data['result'] == 'SUCCESS'
    params = data['params']

    # create peer conection
    pc = RTCPeerConnection()
    pc.addTrack(AudioStreamTrack())
    pc.addTrack(VideoStreamTrack())

    @pc.on('track')
    def on_track(track):
        if track.kind == 'audio':
            consumers.append(asyncio.ensure_future(consume_audio(track)))
        elif track.kind == 'video':
            consumers.append(asyncio.ensure_future(consume_video(track)))

    # connect to websocket and join
    signaling = Signaling()
    await signaling.connect(params)
    await signaling.send({
        'clientid': params['client_id'],
        'cmd': 'register',
        'roomid': params['room_id'],
    })

    if params['is_initiator'] == 'true':
        # send offer
        await pc.setLocalDescription(await pc.createOffer())
        await signaling.send_message(description_to_dict(pc.localDescription))
        print('Please point a browser at %s' % params['room_link'])

    # receive 60s of media
    try:
        await asyncio.wait_for(consume_signaling(signaling, pc, params), timeout=60)
    except asyncio.TimeoutError:
        pass

    # shutdown
    print('Shutting down')
    await signaling.send_message({'type': 'bye'})
    for c in consumers:
        c.cancel()
    await pc.close()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='AppRTC')
    parser.add_argument('room', nargs='?')
    parser.add_argument('--verbose', '-v', action='count')
    args = parser.parse_args()

    if not args.room:
        args.room = ''.join([random.choice('0123456789') for x in range(10)])

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    asyncio.get_event_loop().run_until_complete(join_room(args.room))



from unittest import TestCase

from aiortc.rtcconfiguration import RTCIceServer
from aiortc.rtcicetransport import (RTCIceCandidate, RTCIceGatherer,
                                    RTCIceTransport, connection_kwargs,
                                    parse_stun_turn_uri)

from .utils import run


class ConnectionKwargsTest(TestCase):
    def test_empty(self):
        self.assertEqual(connection_kwargs([]), {})

    def test_stun(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer('stun:stun.l.google.com:19302'),
        ]), {
            'stun_server': ('stun.l.google.com', 19302),
        })

    def test_stun_multiple_servers(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer('stun:stun.l.google.com:19302'),
            RTCIceServer('stun:stun.example.com'),
        ]), {
            'stun_server': ('stun.l.google.com', 19302),
        })

    def test_stun_multiple_urls(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer([
                'stun:stun1.l.google.com:19302',
                'stun:stun2.l.google.com:19302',
            ]),
        ]), {
            'stun_server': ('stun1.l.google.com', 19302),
        })

    def test_turn(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer('turn:turn.example.com'),
        ]), {
            'turn_password': None,
            'turn_server': ('turn.example.com', 3478),
            'turn_username': None,
        })

    def test_turn_multiple_servers(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer('turn:turn.example.com'),
            RTCIceServer('turn:turn.example.net'),
        ]), {
            'turn_password': None,
            'turn_server': ('turn.example.com', 3478),
            'turn_username': None,
        })

    def test_turn_multiple_urls(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer([
                'turn:turn1.example.com',
                'turn:turn2.example.com',
            ])
        ]), {
            'turn_password': None,
            'turn_server': ('turn1.example.com', 3478),
            'turn_username': None,
        })

    def test_turn_over_tcp(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer('turn:turn.example.com?transport=tcp'),
        ]), {})

    def test_turn_with_password(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer(
                urls='turn:turn.example.com',
                username='foo',
                credential='bar'
            ),
        ]), {
            'turn_password': 'bar',
            'turn_server': ('turn.example.com', 3478),
            'turn_username': 'foo',
        })

    def test_turn_with_token(self):
        self.assertEqual(connection_kwargs([
            RTCIceServer(
                urls='turn:turn.example.com',
                username='foo',
                credential='bar',
                credentialType='token',
            ),
        ]), {})


class ParseStunTurnUriTest(TestCase):
    def test_invalid_scheme(self):
        with self.assertRaises(ValueError) as cm:
            parse_stun_turn_uri('foo')
        self.assertEqual(str(cm.exception), 'malformed uri: invalid scheme')

    def test_invalid_uri(self):
        with self.assertRaises(ValueError) as cm:
            parse_stun_turn_uri('stun')
        self.assertEqual(str(cm.exception), 'malformed uri')

    def test_stun(self):
        uri = parse_stun_turn_uri('stun:stun.services.mozilla.com')
        self.assertEqual(uri, {
            'host': 'stun.services.mozilla.com',
            'port': 3478,
            'scheme': 'stun',
        })

    def test_stuns(self):
        uri = parse_stun_turn_uri('stuns:stun.services.mozilla.com')
        self.assertEqual(uri, {
            'host': 'stun.services.mozilla.com',
            'port': 5349,
            'scheme': 'stuns',
        })

    def test_stun_with_port(self):
        uri = parse_stun_turn_uri('stun:stun.l.google.com:19302')
        self.assertEqual(uri, {
            'host': 'stun.l.google.com',
            'port': 19302,
            'scheme': 'stun',
        })

    def test_turn(self):
        uri = parse_stun_turn_uri('turn:1.2.3.4')
        self.assertEqual(uri, {
            'host': '1.2.3.4',
            'port': 3478,
            'scheme': 'turn',
            'transport': 'udp',
        })

    def test_turn_with_port_and_transport(self):
        uri = parse_stun_turn_uri('turn:1.2.3.4:3478?transport=tcp')
        self.assertEqual(uri, {
            'host': '1.2.3.4',
            'port': 3478,
            'scheme': 'turn',
            'transport': 'tcp',
        })

    def test_turns(self):
        uri = parse_stun_turn_uri('turns:1.2.3.4')
        self.assertEqual(uri, {
            'host': '1.2.3.4',
            'port': 5349,
            'scheme': 'turns',
            'transport': 'tcp',
        })

    def test_turns_with_port_and_transport(self):
        uri = parse_stun_turn_uri('turns:1.2.3.4:1234?transport=tcp')
        self.assertEqual(uri, {
            'host': '1.2.3.4',
            'port': 1234,
            'scheme': 'turns',
            'transport': 'tcp',
        })


class RTCIceGathererTest(TestCase):
    def test_construct(self):
        gatherer = RTCIceGatherer()
        self.assertEqual(gatherer.state, 'new')
        self.assertEqual(gatherer.getLocalCandidates(), [])
        run(gatherer.gather())
        self.assertTrue(len(gatherer.getLocalCandidates()) > 0)

    def test_default_ice_servers(self):
        self.assertEqual(RTCIceGatherer.getDefaultIceServers(), [
            RTCIceServer(urls='stun:stun.l.google.com:19302')
        ])


class RTCIceTransportTest(TestCase):
    def test_construct(self):
        gatherer = RTCIceGatherer()
        connection = RTCIceTransport(gatherer)
        self.assertEqual(connection.state, 'new')
        self.assertEqual(connection.getRemoteCandidates(), [])

        candidate = RTCIceCandidate(
            component=1,
            foundation='0',
            ip='192.168.99.7',
            port=33543,
            priority=2122252543,
            protocol='UDP',
            type='host')

        # add candidate
        connection.addRemoteCandidate(candidate)
        self.assertEqual(connection.getRemoteCandidates(), [candidate])

        # end-of-candidates
        connection.addRemoteCandidate(None)
        self.assertEqual(connection.getRemoteCandidates(), [candidate])



import asyncio
import math

from pyee import EventEmitter


class AudioFrame:
    """
    Audio frame, 16-bit PCM.
    """
    def __init__(self, channels, data, sample_rate):
        self.channels = channels
        self.data = data
        self.sample_rate = sample_rate
        self.sample_width = 2


class VideoFrame:
    """
    Video frame in YUV420 format.
    """
    def __init__(self, width, height, data=None):
        self.height = height
        self.width = width
        if data is None:
            self.data = b'\x00' * math.ceil(width * height * 12 / 8)
        else:
            self.data = data


class MediaStreamTrack(EventEmitter):
    pass


class AudioStreamTrack(MediaStreamTrack):
    """
    The base implementation just reads silence.

    Subclass it to provide a useful implementation.
    """
    kind = 'audio'

    async def recv(self):
        await asyncio.sleep(0.02)
        return AudioFrame(channels=1, data=b'\x00\x00' * 160, sample_rate=8000)


class VideoStreamTrack(MediaStreamTrack):
    """
    The base implementation just reads a green frame.

    Subclass it to provide a useful implementation.
    """
    kind = 'video'

    async def recv(self):
        await asyncio.sleep(0.02)
        return VideoFrame(width=320, height=240, data=b'\x00' * 115200)



MAX_MISORDER = 100
MAX_DROPOUT = 3000


class JitterFrame:
    def __init__(self, payload, sequence_number, timestamp):
        self.payload = payload
        self.sequence_number = sequence_number
        self.timestamp = timestamp

    def __repr__(self):
        return 'JitterFrame(seq=%d, ts=%d)' % (self.sequence_number, self.timestamp)


class JitterBuffer:
    def __init__(self, capacity):
        self._capacity = capacity
        self._frames = [None for i in range(capacity)]
        self._head = 0
        self._origin = None

    @property
    def capacity(self):
        return self._capacity

    def add(self, payload, sequence_number, timestamp):
        if self._origin is None:
            self._origin = sequence_number
        elif sequence_number <= self._origin - MAX_MISORDER:
            self.__reset()
            self._origin = sequence_number
        elif sequence_number < self._origin:
            return

        delta = sequence_number - self._origin
        if delta >= self._capacity:
            if delta > MAX_DROPOUT:
                self.__reset()
                self._origin = sequence_number
                delta = 0
            else:
                return

        pos = (self._head + delta) % self._capacity
        self._frames[pos] = JitterFrame(payload=payload,
                                        sequence_number=sequence_number,
                                        timestamp=timestamp)

    def peek(self, offset):
        if offset >= self._capacity:
            raise IndexError('Cannot peek at offset %d, capacity is %d' % (offset, self._capacity))
        pos = (self._head + offset) % self._capacity
        return self._frames[pos]

    def remove(self, count):
        assert count <= self._capacity
        frames = [None for i in range(count)]
        for i in range(count):
            frames[i] = self._frames[self._head]
            self._frames[self._head] = None
            self._head = (self._head + 1) % self._capacity
            self._origin += 1
        return frames

    def __reset(self):
        self._head = 0
        self._origin = None

        for i in range(self._capacity):
            self._frames[i] = None



from unittest import TestCase

from aiortc.jitterbuffer import JitterBuffer


class JitterBufferTest(TestCase):
    def test_create(self):
        jbuffer = JitterBuffer(capacity=2)
        self.assertEqual(jbuffer._frames, [None, None])
        self.assertEqual(jbuffer._origin, None)

        jbuffer = JitterBuffer(capacity=4)
        self.assertEqual(jbuffer._frames, [None, None, None, None])
        self.assertEqual(jbuffer._origin, None)

    def test_add_ordered(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)
        jbuffer.add(b'0002', sequence_number=2, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)
        jbuffer.add(b'0003', sequence_number=3, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)
        jbuffer.add(b'0004', sequence_number=4, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)

        self.assertIsNotNone(jbuffer._frames[0])
        self.assertEqual(jbuffer._frames[0].payload, b'0001')
        self.assertEqual(jbuffer._frames[0].sequence_number, 1)
        self.assertIsNotNone(jbuffer._frames[1])
        self.assertEqual(jbuffer._frames[1].payload, b'0002')
        self.assertEqual(jbuffer._frames[1].sequence_number, 2)
        self.assertIsNotNone(jbuffer._frames[2])
        self.assertEqual(jbuffer._frames[2].payload, b'0003')
        self.assertEqual(jbuffer._frames[2].sequence_number, 3)
        self.assertIsNotNone(jbuffer._frames[3])
        self.assertEqual(jbuffer._frames[3].payload, b'0004')
        self.assertEqual(jbuffer._frames[3].sequence_number, 4)

    def test_add_unordered(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)
        jbuffer.add(b'0003', sequence_number=3, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)
        jbuffer.add(b'0002', sequence_number=2, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)

        self.assertIsNotNone(jbuffer._frames[0])
        self.assertEqual(jbuffer._frames[0].payload, b'0001')
        self.assertEqual(jbuffer._frames[0].sequence_number, 1)
        self.assertIsNotNone(jbuffer._frames[1])
        self.assertEqual(jbuffer._frames[1].payload, b'0002')
        self.assertEqual(jbuffer._frames[1].sequence_number, 2)
        self.assertIsNotNone(jbuffer._frames[2])
        self.assertEqual(jbuffer._frames[2].payload, b'0003')
        self.assertEqual(jbuffer._frames[2].sequence_number, 3)
        self.assertIsNone(jbuffer._frames[3])

    def test_add_seq_too_low_drop(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'0002', sequence_number=2, timestamp=1234)
        self.assertEqual(jbuffer._origin, 2)
        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        self.assertEqual(jbuffer._origin, 2)

        self.assertIsNotNone(jbuffer._frames[0])
        self.assertEqual(jbuffer._frames[0].payload, b'0002')
        self.assertEqual(jbuffer._frames[0].sequence_number, 2)
        self.assertIsNone(jbuffer._frames[1])
        self.assertIsNone(jbuffer._frames[2])
        self.assertIsNone(jbuffer._frames[3])

    def test_add_seq_too_low_reset(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'2000', sequence_number=2000, timestamp=1234)
        self.assertEqual(jbuffer._origin, 2000)
        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)

        self.assertIsNotNone(jbuffer._frames[0])
        self.assertEqual(jbuffer._frames[0].payload, b'0001')
        self.assertEqual(jbuffer._frames[0].sequence_number, 1)
        self.assertIsNone(jbuffer._frames[1])
        self.assertIsNone(jbuffer._frames[2])
        self.assertIsNone(jbuffer._frames[3])

    def test_add_seq_too_high_drop(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)
        jbuffer.add(b'0005', sequence_number=5, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)

        self.assertIsNotNone(jbuffer._frames[0])
        self.assertEqual(jbuffer._frames[0].payload, b'0001')
        self.assertEqual(jbuffer._frames[0].sequence_number, 1)
        self.assertIsNone(jbuffer._frames[1])
        self.assertIsNone(jbuffer._frames[2])
        self.assertIsNone(jbuffer._frames[3])

    def test_add_seq_too_high_reset(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        self.assertEqual(jbuffer._origin, 1)
        jbuffer.add(b'3002', sequence_number=3002, timestamp=1234)
        self.assertEqual(jbuffer._origin, 3002)

        self.assertIsNotNone(jbuffer._frames[0])
        self.assertEqual(jbuffer._frames[0].payload, b'3002')
        self.assertEqual(jbuffer._frames[0].sequence_number, 3002)
        self.assertIsNone(jbuffer._frames[1])
        self.assertIsNone(jbuffer._frames[2])
        self.assertIsNone(jbuffer._frames[3])

    def test_peek(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        jbuffer.add(b'0002', sequence_number=2, timestamp=1234)
        jbuffer.add(b'0004', sequence_number=4, timestamp=1234)

        frame = jbuffer.peek(0)
        self.assertIsNotNone(frame)
        self.assertEqual(frame.sequence_number, 1)
        self.assertEqual(repr(frame), 'JitterFrame(seq=1, ts=1234)')

        frame = jbuffer.peek(1)
        self.assertIsNotNone(frame)
        self.assertEqual(frame.sequence_number, 2)

        frame = jbuffer.peek(2)
        self.assertIsNone(frame)

        frame = jbuffer.peek(3)
        self.assertIsNotNone(frame)
        self.assertEqual(frame.sequence_number, 4)

        with self.assertRaises(IndexError) as cm:
            jbuffer.peek(4)
        self.assertEqual(str(cm.exception), 'Cannot peek at offset 4, capacity is 4')

    def test_remove(self):
        jbuffer = JitterBuffer(capacity=4)

        jbuffer.add(b'0001', sequence_number=1, timestamp=1234)
        jbuffer.add(b'0002', sequence_number=2, timestamp=1234)
        jbuffer.add(b'0003', sequence_number=3, timestamp=1234)
        jbuffer.add(b'0004', sequence_number=4, timestamp=1234)

        # remove 1 frame
        frames = jbuffer.remove(1)
        self.assertEqual(len(frames), 1)
        self.assertEqual(frames[0].sequence_number, 1)

        # check buffer
        self.assertEqual(jbuffer._head, 1)
        self.assertEqual(jbuffer._origin, 2)
        self.assertIsNone(jbuffer._frames[0])
        self.assertIsNotNone(jbuffer._frames[1])
        self.assertIsNotNone(jbuffer._frames[2])
        self.assertIsNotNone(jbuffer._frames[3])

        # remove 2 frames
        frames = jbuffer.remove(2)
        self.assertEqual(len(frames), 2)
        self.assertEqual(frames[0].sequence_number, 2)
        self.assertEqual(frames[1].sequence_number, 3)

        # check buffer
        self.assertEqual(jbuffer._head, 3)
        self.assertEqual(jbuffer._origin, 4)
        self.assertIsNone(jbuffer._frames[0])
        self.assertIsNone(jbuffer._frames[1])
        self.assertIsNone(jbuffer._frames[2])
        self.assertIsNotNone(jbuffer._frames[3])



import math
import multiprocessing
from struct import pack, unpack

from ..mediastreams import VideoFrame
from ._vpx import ffi, lib

PACKET_MAX = 1300 - 1
MAX_FRAME_RATE = 30


def number_of_threads(pixels, cpus):
    if pixels >= 1920 * 1080 and cpus > 8:
        return 8
    elif pixels > 1280 * 960 and cpus >= 6:
        return 3
    elif pixels > 640 * 480 and cpus >= 3:
        return 2
    else:
        return 1


class VpxPayloadDescriptor:
    props = ['partition_start', 'partition_id', 'picture_id']

    def __init__(self, partition_start, partition_id, picture_id=None,
                 tl0picidx=None, tid=None, keyidx=None):
        self.partition_start = partition_start
        self.partition_id = partition_id
        self.picture_id = picture_id
        self.tl0picidx = tl0picidx
        self.tid = tid
        self.keyidx = keyidx

    def __bytes__(self):
        octet = (self.partition_start << 4) | self.partition_id

        ext_octet = 0
        if self.picture_id is not None:
            ext_octet |= 1 << 7
        if self.tl0picidx is not None:
            ext_octet |= 1 << 6
        if self.tid is not None:
            ext_octet |= 1 << 5
        if self.keyidx is not None:
            ext_octet |= 1 << 4

        if ext_octet:
            data = pack('!BB', (1 << 7) | octet, ext_octet)
            if self.picture_id is not None:
                if self.picture_id < 128:
                    data += pack('!B', self.picture_id)
                else:
                    data += pack('!H', (1 << 15) | self.picture_id)
            if self.tl0picidx is not None:
                data += pack('!B', self.tl0picidx)
            if self.tid is not None or self.keyidx is not None:
                t_k = 0
                if self.tid is not None:
                    t_k |= (self.tid[0] << 6) | (self.tid[1] << 5)
                if self.keyidx is not None:
                    t_k |= self.keyidx
                data += pack('!B', t_k)
        else:
            data = pack('!B', octet)

        return data

    def __repr__(self):
        return 'VpxPayloadDescriptor(S=%d, PID=%d, pic_id=%s)' % (
            self.partition_start, self.partition_id, self.picture_id)

    @classmethod
    def parse(cls, data):
        # first byte
        octet = data[0]
        extended = octet >> 7
        partition_start = (octet >> 4) & 1
        partition_id = octet & 0xf
        picture_id = None
        tl0picidx = None
        tid = None
        keyidx = None
        pos = 1

        # extended control bits
        if extended:
            octet = data[pos]
            ext_I = (octet >> 7) & 1
            ext_L = (octet >> 6) & 1
            ext_T = (octet >> 5) & 1
            ext_K = (octet >> 4) & 1
            pos += 1

            # picture id
            if ext_I:
                if data[pos] & 0x80:
                    picture_id = unpack('!H', data[pos:pos+2])[0] & 0x7fff
                    pos += 2
                else:
                    picture_id = data[pos]
                    pos += 1

            # unused
            if ext_L:
                tl0picidx = unpack('!B', data[pos:pos+1])[0]
                pos += 1
            if ext_T or ext_K:
                t_k = unpack('!B', data[pos:pos+1])[0]
                if ext_T:
                    tid = (
                        (t_k >> 6) & 3,
                        (t_k >> 5) & 1
                    )
                if ext_K:
                    keyidx = t_k & 0x1f
                pos += 1

        obj = cls(partition_start=partition_start, partition_id=partition_id, picture_id=picture_id,
                  tl0picidx=tl0picidx, tid=tid, keyidx=keyidx)
        return obj, data[pos:]


def _vpx_assert(err):
    if err != lib.VPX_CODEC_OK:
        reason = ffi.string(lib.vpx_codec_err_to_string(err))
        raise Exception('libvpx error: ' + reason.decode('utf8'))


class VpxDecoder:
    def __init__(self):
        self.codec = ffi.new('vpx_codec_ctx_t *')
        _vpx_assert(lib.vpx_codec_dec_init(self.codec, lib.vpx_codec_vp8_dx(), ffi.NULL, 0))

    def __del__(self):
        lib.vpx_codec_destroy(self.codec)

    def decode(self, payloads):
        data = b''
        for payload in payloads:
            if payload:
                vpx_descriptor, rest = VpxPayloadDescriptor.parse(payload)
                data += rest

        frames = []
        result = lib.vpx_codec_decode(self.codec, data, len(data), ffi.NULL, lib.VPX_DL_REALTIME)
        if result == lib.VPX_CODEC_OK:
            it = ffi.new('vpx_codec_iter_t *')
            while True:
                img = lib.vpx_codec_get_frame(self.codec, it)
                if not img:
                    break
                assert img.fmt == lib.VPX_IMG_FMT_I420

                o_buf = bytearray(math.ceil(img.d_w * img.d_h * 12 / 8))
                o_pos = 0
                for p in range(3):
                    i_stride = img.stride[p]
                    i_buf = ffi.buffer(img.planes[p], i_stride * img.d_h)
                    i_pos = 0

                    div = p and 2 or 1
                    o_stride = img.d_w // div
                    for r in range(0, img.d_h // div):
                        o_buf[o_pos:o_pos + o_stride] = i_buf[i_pos:i_pos + o_stride]
                        i_pos += i_stride
                        o_pos += o_stride

                frames.append(VideoFrame(width=img.d_w, height=img.d_h, data=bytes(o_buf)))

        return frames


class VpxEncoder:
    timestamp_increment = 3000

    def __init__(self):
        self.cx = lib.vpx_codec_vp8_cx()

        self.cfg = ffi.new('vpx_codec_enc_cfg_t *')
        lib.vpx_codec_enc_config_default(self.cx, self.cfg, 0)

        self.codec = None
        self.timestamp = 0

    def __del__(self):
        if self.codec:
            lib.vpx_codec_destroy(self.codec)

    def encode(self, frame):
        image = ffi.new('vpx_image_t *')

        lib.vpx_img_wrap(image, lib.VPX_IMG_FMT_I420,
                         frame.width, frame.height, 1, frame.data)

        if not self.codec:
            self.codec = ffi.new('vpx_codec_ctx_t *')
            self.cfg.g_timebase.num = 1
            self.cfg.g_timebase.den = 90000
            self.cfg.g_threads = number_of_threads(frame.width * frame.height,
                                                   multiprocessing.cpu_count())
            self.cfg.g_w = frame.width
            self.cfg.g_h = frame.height
            self.cfg.rc_end_usage = lib.VPX_CBR
            self.cfg.rc_min_quantizer = 2
            self.cfg.rc_max_quantizer = 56
            self.cfg.rc_undershoot_pct = 100
            self.cfg.rc_overshoot_pct = 15
            self.cfg.rc_buf_initial_sz = 500
            self.cfg.rc_buf_optimal_sz = 600
            self.cfg.rc_buf_sz = 1000
            self.cfg.kf_mode = lib.VPX_KF_AUTO
            self.cfg.kf_max_dist = 3000
            _vpx_assert(lib.vpx_codec_enc_init(self.codec, self.cx, self.cfg, 0))
        elif frame.width != self.cfg.g_w or frame.height != self.cfg.g_h:
            self.cfg.g_w = frame.width
            self.cfg.g_h = frame.height
            _vpx_assert(lib.vpx_codec_enc_config_set(self.codec, self.cfg))

        duration = 90000 // MAX_FRAME_RATE
        _vpx_assert(lib.vpx_codec_encode(
            self.codec, image, self.timestamp, duration,  0, lib.VPX_DL_REALTIME))
        self.timestamp += duration

        it = ffi.new('vpx_codec_iter_t *')
        payloads = []
        while True:
            pkt = lib.vpx_codec_get_cx_data(self.codec, it)
            if not pkt:
                break
            if pkt and pkt.kind == lib.VPX_CODEC_CX_FRAME_PKT:
                buf = ffi.buffer(pkt.data.frame.buf, pkt.data.frame.sz)
                descr = VpxPayloadDescriptor(partition_start=1, partition_id=0)
                for pos in range(0, len(buf), PACKET_MAX):
                    data = buf[pos:pos + PACKET_MAX]
                    payloads.append(bytes(descr) + data)
                    descr.partition_start = 0
        return payloads



from .exceptions import InvalidAccessError, InvalidStateError  # noqa
from .mediastreams import AudioStreamTrack, VideoStreamTrack  # noqa
from .rtcconfiguration import RTCConfiguration, RTCIceServer  # noqa
from .rtcdatachannel import RTCDataChannel, RTCDataChannelParameters  # noqa
from .rtcdtlstransport import (RTCCertificate, RTCDtlsFingerprint,  # noqa
                               RTCDtlsParameters, RTCDtlsTransport)
from .rtcicetransport import (RTCIceCandidate, RTCIceGatherer,  # noqa
                              RTCIceParameters, RTCIceTransport)
from .rtcpeerconnection import RTCPeerConnection  # noqa
from .rtcrtpparameters import (RTCRtcpParameters,  # noqa
                               RTCRtpCodecParameters, RTCRtpParameters)
from .rtcrtpreceiver import RTCRtpReceiver  # noqa
from .rtcrtpsender import RTCRtpSender  # noqa
from .rtcrtptransceiver import RTCRtpTransceiver  # noqa
from .rtcsctptransport import RTCSctpCapabilities, RTCSctpTransport  # noqa
from .rtcsessiondescription import RTCSessionDescription  # noqa



import asyncio
import base64
import binascii
import datetime
import enum
import logging
import os
import struct

import attr
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.bindings.openssl.binding import Binding
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import (Encoding,
                                                          NoEncryption,
                                                          PrivateFormat)
from OpenSSL import crypto
from pyee import EventEmitter
from pylibsrtp import Policy, Session

from .rtp import RtcpPacket, RtpPacket, is_rtcp
from .utils import first_completed

binding = Binding()
binding.init_static_locks()
ffi = binding.ffi
lib = binding.lib

SRTP_KEY_LEN = 16
SRTP_SALT_LEN = 14

logger = logging.getLogger('dtls')


class DtlsError(Exception):
    pass


def _openssl_assert(ok):
    if not ok:
        raise DtlsError('OpenSSL call failed')


def certificate_digest(x509):
    digest = lib.EVP_get_digestbyname(b'SHA256')
    _openssl_assert(digest != ffi.NULL)

    result_buffer = ffi.new('unsigned char[]', lib.EVP_MAX_MD_SIZE)
    result_length = ffi.new('unsigned int[]', 1)
    result_length[0] = len(result_buffer)

    digest_result = lib.X509_digest(x509, digest, result_buffer, result_length)
    assert digest_result == 1

    return b":".join([
        base64.b16encode(ch).upper() for ch
        in ffi.buffer(result_buffer, result_length[0])]).decode('ascii')


def generate_key():
    key = ec.generate_private_key(ec.SECP256R1(), default_backend())
    key_pem = key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=NoEncryption())
    return crypto.load_privatekey(crypto.FILETYPE_PEM, key_pem)


def generate_certificate(key):
    cert = crypto.X509()
    cert.get_subject().CN = binascii.hexlify(os.urandom(16)).decode('ascii')
    cert.gmtime_adj_notBefore(-86400)
    cert.gmtime_adj_notAfter(30 * 86400)
    cert.set_version(2)
    cert.set_serial_number(struct.unpack('!L', os.urandom(4))[0])
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(key)
    cert.sign(key, 'sha256')
    return cert


def get_srtp_key_salt(src, idx):
    key_start = idx * SRTP_KEY_LEN
    salt_start = 2 * SRTP_KEY_LEN + idx * SRTP_SALT_LEN
    return (
        src[key_start:key_start + SRTP_KEY_LEN] +
        src[salt_start:salt_start + SRTP_SALT_LEN]
    )


@ffi.callback('int(int, X509_STORE_CTX *)')
def verify_callback(x, y):
    return 1


def create_ssl_context(certificate):
    if hasattr(lib, 'DTLS_method'):
        # openssl >= 1.0.2
        method = lib.DTLS_method
    else:  # pragma: no cover
        # openssl < 1.0.2
        method = lib.DTLSv1_method
    ctx = lib.SSL_CTX_new(method())
    ctx = ffi.gc(ctx, lib.SSL_CTX_free)

    lib.SSL_CTX_set_verify(ctx, lib.SSL_VERIFY_PEER | lib.SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                           verify_callback)

    _openssl_assert(lib.SSL_CTX_use_certificate(ctx, certificate._cert._x509) == 1)
    _openssl_assert(lib.SSL_CTX_use_PrivateKey(ctx, certificate._key._pkey) == 1)
    _openssl_assert(lib.SSL_CTX_set_cipher_list(ctx, b'HIGH:!CAMELLIA:!aNULL') == 1)
    _openssl_assert(lib.SSL_CTX_set_tlsext_use_srtp(ctx, b'SRTP_AES128_CM_SHA1_80') == 0)
    _openssl_assert(lib.SSL_CTX_set_read_ahead(ctx, 1) == 0)
    return ctx


class Channel:
    def __init__(self, closed, queue, send):
        self.closed = closed
        self.queue = queue
        self.send = send

    async def recv(self):
        data = await first_completed(self.queue.get(), self.closed.wait())
        if data is True:
            raise ConnectionError
        return data


class State(enum.Enum):
    NEW = 0
    CONNECTING = 1
    CONNECTED = 2
    CLOSED = 3
    FAILED = 4


class RTCCertificate:
    """
    The :class:`RTCCertificate` interface enables the certificates used by an
    :class:`RTCDtlsTransport`.

    To generate a certificate and the corresponding private key use :func:`generateCertificate`.
    """
    def __init__(self, key, cert):
        self._key = key
        self._cert = cert

    @property
    def expires(self):
        """
        The date and time after which the certificate will be considered invalid.
        """
        not_after = self._cert.get_notAfter().decode('ascii')
        return datetime.datetime.strptime(not_after, '%Y%m%d%H%M%SZ').replace(
            tzinfo=datetime.timezone.utc)

    def getFingerprints(self):
        """
        Returns the list of certificate fingerprints, one of which is computed
        with the digest algorithm used in the certificate signature.
        """
        return [
            RTCDtlsFingerprint(algorithm='sha-256', value=certificate_digest(self._cert._x509))
        ]

    @classmethod
    def generateCertificate(cls):
        """
        Create and return an X.509 certificate and corresponding private key.

        :rtype: RTCCertificate
        """
        key = generate_key()
        cert = generate_certificate(key)
        return cls(key=key, cert=cert)


@attr.s
class RTCDtlsFingerprint:
    """
    The :class:`RTCDtlsFingerprint` dictionary includes the hash function
    algorithm and certificate fingerprint.
    """
    algorithm = attr.ib()
    "The hash function name, for instance `'sha-256'`."

    value = attr.ib()
    "The fingerprint value."


@attr.s
class RTCDtlsParameters:
    """
    The :class:`RTCDtlsParameters` dictionary includes information relating to
    DTLS configuration.
    """
    fingerprints = attr.ib(default=attr.Factory(list))
    "List of :class:`RTCDtlsFingerprint`, one fingerprint for each certificate."

    role = attr.ib(default='auto')
    "The DTLS role, with a default of auto."


class RtpRouter:
    def __init__(self):
        self.ssrc_table = {}

    def register(self, receiver, parameters):
        if parameters.rtcp.ssrc:
            self.ssrc_table[parameters.rtcp.ssrc] = receiver

    def route(self, ssrc):
        return self.ssrc_table.get(ssrc)


class RTCDtlsTransport(EventEmitter):
    """
    The :class:`RTCDtlsTransport` object includes information relating to
    Datagram Transport Layer Security (DTLS) transport.

    :param: transport: An :class:`RTCIceTransport`.
    :param: certificates: A list of :class:`RTCCertificate` (only one is allowed currently).
    """
    def __init__(self, transport, certificates):
        assert len(certificates) == 1
        certificate = certificates[0]

        super().__init__()
        self.closed = asyncio.Event()
        self.encrypted = False
        self._role = 'auto'
        self._rtp_router = RtpRouter()
        self._start = None
        self._state = State.NEW
        self._transport = transport

        self.data_queue = asyncio.Queue()
        self.data = Channel(
            closed=self.closed,
            queue=self.data_queue,
            send=self._send_data)

        # SSL init
        self.__ctx = create_ssl_context(certificate)

        ssl = lib.SSL_new(self.__ctx)
        self.ssl = ffi.gc(ssl, lib.SSL_free)

        self.read_bio = lib.BIO_new(lib.BIO_s_mem())
        self.read_cdata = ffi.new('char[]', 1500)
        self.write_bio = lib.BIO_new(lib.BIO_s_mem())
        self.write_cdata = ffi.new('char[]', 1500)
        lib.SSL_set_bio(self.ssl, self.read_bio, self.write_bio)

        self.__local_parameters = RTCDtlsParameters(fingerprints=certificate.getFingerprints())

    @property
    def state(self):
        """
        The current state of the DTLS transport.
        """
        return str(self._state)[6:].lower()

    @property
    def transport(self):
        """
        The associated :class:`RTCIceTransport` instance.
        """
        return self._transport

    def getLocalParameters(self):
        """
        Get the local parameters of the DTLS transport.

        :rtype: :class:`RTCDtlsParameters`
        """
        return self.__local_parameters

    async def start(self, remoteParameters):
        """
        Start DTLS transport negotiation with the parameters of the remote
        DTLS transport.

        :param: remoteParameters: An :class:`RTCDtlsParameters`.
        """
        assert self._state not in [State.CLOSED, State.FAILED]
        assert len(remoteParameters.fingerprints)

        # handle the case where start is already in progress
        if self._start is not None:
            return await self._start.wait()
        self._start = asyncio.Event()

        if self.transport.role == 'controlling':
            self._role = 'server'
            lib.SSL_set_accept_state(self.ssl)
        else:
            self._role = 'client'
            lib.SSL_set_connect_state(self.ssl)

        self._set_state(State.CONNECTING)
        while not self.encrypted:
            result = lib.SSL_do_handshake(self.ssl)
            await self._write_ssl()

            if result > 0:
                self.encrypted = True
                break

            error = lib.SSL_get_error(self.ssl, result)
            if error == lib.SSL_ERROR_WANT_READ:
                await self._recv_next()
            else:
                self._set_state(State.FAILED)
                raise DtlsError('DTLS handshake failed (error %d)' % error)

        # check remote fingerprint
        x509 = lib.SSL_get_peer_certificate(self.ssl)
        remote_fingerprint = certificate_digest(x509)
        fingerprint_is_valid = False
        for f in remoteParameters.fingerprints:
            if f.algorithm == 'sha-256' and f.value.lower() == remote_fingerprint.lower():
                fingerprint_is_valid = True
                break
        if not fingerprint_is_valid:
            self._set_state(State.FAILED)
            raise DtlsError('DTLS fingerprint does not match')

        # generate keying material
        buf = ffi.new('unsigned char[]', 2 * (SRTP_KEY_LEN + SRTP_SALT_LEN))
        extractor = b'EXTRACTOR-dtls_srtp'
        _openssl_assert(lib.SSL_export_keying_material(
            self.ssl, buf, len(buf), extractor, len(extractor), ffi.NULL, 0, 0) == 1)

        view = ffi.buffer(buf)
        if self._role == 'server':
            srtp_tx_key = get_srtp_key_salt(view, 1)
            srtp_rx_key = get_srtp_key_salt(view, 0)
        else:
            srtp_tx_key = get_srtp_key_salt(view, 0)
            srtp_rx_key = get_srtp_key_salt(view, 1)

        rx_policy = Policy(key=srtp_rx_key, ssrc_type=Policy.SSRC_ANY_INBOUND)
        self._rx_srtp = Session(rx_policy)
        tx_policy = Policy(key=srtp_tx_key, ssrc_type=Policy.SSRC_ANY_OUTBOUND)
        self._tx_srtp = Session(tx_policy)

        # start data pump
        self.__log_debug('- DTLS handshake complete')
        self._set_state(State.CONNECTED)
        asyncio.ensure_future(self.__run())
        self._start.set()

    async def stop(self):
        """
        Stop and close the DTLS transport.
        """
        if self._state in [State.CONNECTING, State.CONNECTED]:
            lib.SSL_shutdown(self.ssl)
            await self._write_ssl()
            self.__log_debug('- DTLS shutdown complete')
            self.closed.set()

    async def __run(self):
        try:
            while True:
                await self._recv_next()
        except ConnectionError:
            pass
        finally:
            self._set_state(State.CLOSED)
            self.closed.set()

    async def _recv_next(self):
        # get timeout
        ptv_sec = ffi.new('time_t *')
        ptv_usec = ffi.new('long *')
        if lib.Cryptography_DTLSv1_get_timeout(self.ssl, ptv_sec, ptv_usec):
            timeout = ptv_sec[0] + (ptv_usec[0] / 1000000)
        else:
            timeout = None

        try:
            data = await first_completed(self.transport._connection.recv(), self.closed.wait(),
                                         timeout=timeout)
        except TimeoutError:
            self.__log_debug('x DTLS handling timeout')
            lib.DTLSv1_handle_timeout(self.ssl)
            await self._write_ssl()
            return

        if data is True:
            # session was closed
            raise ConnectionError

        first_byte = data[0]
        if first_byte > 19 and first_byte < 64:
            # DTLS
            lib.BIO_write(self.read_bio, data, len(data))
            result = lib.SSL_read(self.ssl, self.read_cdata, len(self.read_cdata))
            await self._write_ssl()
            if result == 0:
                self.__log_debug('- DTLS shutdown by remote party')
                raise ConnectionError
            elif result > 0:
                await self.data_queue.put(ffi.buffer(self.read_cdata)[0:result])
        elif first_byte > 127 and first_byte < 192:
            # SRTP / SRTCP
            if is_rtcp(data):
                data = self._rx_srtp.unprotect_rtcp(data)
                packets = RtcpPacket.parse(data)
                for packet in packets:
                    receiver = None
                    if hasattr(packet, 'ssrc'):
                        # SR and RR
                        receiver = self._rtp_router.route(packet.ssrc)
                    elif getattr(packet, 'chunks', None):
                        # SDES
                        receiver = self._rtp_router.route(packet.chunks[0].ssrc)
                    elif getattr(packet, 'sources', None):
                        # BYE
                        receiver = self._rtp_router.route(packet.sources[0])
                    if receiver is not None:
                        await receiver._handle_rtcp_packet(packet)
            else:
                data = self._rx_srtp.unprotect(data)
                packet = RtpPacket.parse(data)
                receiver = self._rtp_router.route(packet.ssrc)
                if receiver is not None:
                    await receiver._handle_rtp_packet(packet)

    def _register_rtp_receiver(self, receiver, parameters):
        self._rtp_router.register(receiver, parameters)

    async def _send_data(self, data):
        if self._state != State.CONNECTED:
            raise ConnectionError('Cannot send encrypted data, not connected')

        lib.SSL_write(self.ssl, data, len(data))
        await self._write_ssl()

    async def _send_rtp(self, data):
        if self._state != State.CONNECTED:
            raise ConnectionError('Cannot send encrypted RTP, not connected')

        if is_rtcp(data):
            data = self._tx_srtp.protect_rtcp(data)
        else:
            data = self._tx_srtp.protect(data)
        await self.transport._connection.send(data)

    def _set_state(self, state):
        if state != self._state:
            self.__log_debug('- %s -> %s', self._state, state)
            self._state = state
            self.emit('statechange')

    async def _write_ssl(self):
        """
        Flush outgoing data which OpenSSL put in our BIO to the transport.
        """
        pending = lib.BIO_ctrl_pending(self.write_bio)
        if pending > 0:
            result = lib.BIO_read(self.write_bio, self.write_cdata, len(self.write_cdata))
            await self.transport._connection.send(ffi.buffer(self.write_cdata)[0:result])

    def __log_debug(self, msg, *args):
        logger.debug(self._role + ' ' + msg, *args)



from unittest import TestCase

from aiortc.codecs import PCMU_CODEC
from aiortc.exceptions import InvalidStateError
from aiortc.mediastreams import AudioFrame
from aiortc.rtcrtpparameters import RTCRtpParameters
from aiortc.rtcrtpreceiver import (LossCounter, RemoteStreamTrack,
                                   RTCRtpReceiver)
from aiortc.rtp import RtcpPacket, RtpPacket

from .utils import dummy_dtls_transport_pair, load, run


class ClosedDtlsTransport:
    state = 'closed'


class LossCounterTest(TestCase):
    def test_no_loss(self):
        # receive 10 packets
        counter = LossCounter(0)
        for seq in range(1, 10):
            counter.add(seq)
        self.assertEqual(counter.max_seq, 9)
        self.assertEqual(counter.packets_received, 10)
        self.assertEqual(counter.packets_lost, 0)
        self.assertEqual(counter.fraction_lost, 0)

        # receive 10 more packets
        for seq in range(10, 20):
            counter.add(seq)
        self.assertEqual(counter.max_seq, 19)
        self.assertEqual(counter.packets_received, 20)
        self.assertEqual(counter.packets_lost, 0)
        self.assertEqual(counter.fraction_lost, 0)

    def test_no_loss_cycle(self):
        counter = LossCounter(65530)
        counter.add(65531)
        counter.add(65532)
        counter.add(65533)
        counter.add(65534)
        counter.add(65535)
        counter.add(0)
        counter.add(1)
        counter.add(2)
        counter.add(3)
        self.assertEqual(counter.max_seq, 3)
        self.assertEqual(counter.packets_received, 10)
        self.assertEqual(counter.packets_lost, 0)
        self.assertEqual(counter.fraction_lost, 0)

    def test_with_loss(self):
        # receive 9 packets (one missing)
        counter = LossCounter(0)
        for seq in range(2, 10):
            counter.add(seq)
        self.assertEqual(counter.max_seq, 9)
        self.assertEqual(counter.packets_received, 9)
        self.assertEqual(counter.packets_lost, 1)
        self.assertEqual(counter.fraction_lost, 25)

        # receive 10 more packets
        for seq in range(10, 20):
            counter.add(seq)
        self.assertEqual(counter.max_seq, 19)
        self.assertEqual(counter.packets_received, 19)
        self.assertEqual(counter.packets_lost, 1)
        self.assertEqual(counter.fraction_lost, 0)


class RTCRtpReceiverTest(TestCase):
    def test_connection_error(self):
        """
        Close the underlying transport before the receiver.
        """
        transport, _ = dummy_dtls_transport_pair()

        receiver = RTCRtpReceiver('audio', transport)
        self.assertEqual(receiver.transport, transport)

        run(receiver.receive(RTCRtpParameters(codecs=[PCMU_CODEC])))

        run(transport.close())

    def test_rtp_and_rtcp(self):
        transport, remote = dummy_dtls_transport_pair()

        receiver = RTCRtpReceiver('audio', transport)
        self.assertEqual(receiver.transport, transport)

        receiver._track = RemoteStreamTrack(kind='audio')
        run(receiver.receive(RTCRtpParameters(codecs=[PCMU_CODEC])))

        # receive RTP
        packet = RtpPacket.parse(load('rtp.bin'))
        run(receiver._handle_rtp_packet(packet))

        # receive RTCP
        for packet in RtcpPacket.parse(load('rtcp_sr.bin')):
            run(receiver._handle_rtcp_packet(packet))
        self.assertEqual(sorted(receiver._stats.keys()), [
            'remote-inbound-rtp',
            'remote-outbound-rtp'
        ])

        # check remote track
        frame = run(receiver._track.recv())
        self.assertTrue(isinstance(frame, AudioFrame))

        # shutdown
        run(receiver.stop())

    def test_invalid_dtls_transport_state(self):
        dtlsTransport = ClosedDtlsTransport()
        with self.assertRaises(InvalidStateError):
            RTCRtpReceiver('audio', dtlsTransport)



import audioop

from ..mediastreams import AudioFrame

SAMPLE_RATE = 8000


def mono_8khz(frame):
    data = frame.data

    # resample at 8 kHz
    if frame.sample_rate != SAMPLE_RATE:
        data, _ = audioop.ratecv(
            data,
            frame.sample_width,
            frame.channels,
            frame.sample_rate,
            SAMPLE_RATE,
            None)

    # convert to mono
    if frame.channels == 2:
        data = audioop.tomono(data, frame.sample_width, 1, 1)

    return data


class PcmaDecoder:
    def decode(self, data):
        return AudioFrame(
            channels=1,
            data=audioop.alaw2lin(data, 2),
            sample_rate=8000)


class PcmaEncoder:
    timestamp_increment = 160

    def encode(self, frame):
        return audioop.lin2alaw(mono_8khz(frame), frame.sample_width)


class PcmuDecoder:
    def decode(self, data):
        return AudioFrame(
            channels=1,
            data=audioop.ulaw2lin(data, 2),
            sample_rate=8000)


class PcmuEncoder:
    timestamp_increment = 160

    def encode(self, frame):
        return audioop.lin2ulaw(mono_8khz(frame), frame.sample_width)



from unittest import TestCase

from aiortc.codecs import PCMU_CODEC
from aiortc.exceptions import InvalidStateError
from aiortc.mediastreams import AudioStreamTrack
from aiortc.rtcrtpparameters import RTCRtpParameters
from aiortc.rtcrtpsender import RTCRtpSender

from .utils import dummy_dtls_transport_pair, run


class ClosedDtlsTransport:
    state = 'closed'


class RTCRtpSenderTest(TestCase):
    def test_construct(self):
        transport, _ = dummy_dtls_transport_pair()

        sender = RTCRtpSender('audio', transport)
        self.assertEqual(sender.kind, 'audio')
        self.assertEqual(sender.transport, transport)

    def test_construct_invalid_dtls_transport_state(self):
        transport = ClosedDtlsTransport()

        with self.assertRaises(InvalidStateError):
            RTCRtpSender('audio', transport)

    def test_connection_error(self):
        """
        Close the underlying transport before the sender.
        """
        transport, _ = dummy_dtls_transport_pair()

        sender = RTCRtpSender(AudioStreamTrack(), transport)
        self.assertEqual(sender.kind, 'audio')
        self.assertEqual(sender.transport, transport)

        run(sender.send(RTCRtpParameters(codecs=[PCMU_CODEC])))

        run(transport.close())

    def test_stop(self):
        transport, _ = dummy_dtls_transport_pair()

        sender = RTCRtpSender(AudioStreamTrack(), transport)
        self.assertEqual(sender.kind, 'audio')
        self.assertEqual(sender.transport, transport)

        run(sender.send(RTCRtpParameters(codecs=[PCMU_CODEC])))

        # clean shutdown
        run(sender.stop())

    def test_stop_before_send(self):
        transport, _ = dummy_dtls_transport_pair()
        sender = RTCRtpSender(AudioStreamTrack(), transport)
        run(sender.stop())



import asyncio
import logging
import random

from .codecs import get_encoder
from .exceptions import InvalidStateError
from .rtp import (RtcpByePacket, RtcpSdesPacket, RtcpSenderInfo,
                  RtcpSourceInfo, RtcpSrPacket, RtpPacket, datetime_to_ntp,
                  seq_plus_one, utcnow)
from .utils import first_completed, random32

logger = logging.getLogger('rtp')


class RTCRtpSender:
    """
    The :class:`RTCRtpSender` interface provides the ability to control and
    obtain details about how a particular :class:`MediaStreamTrack` is encoded
    and sent to a remote peer.

    :param: trackOrKind: Either a :class:`MediaStreamTrack` instance or a
                         media kind (`'audio'` or `'video'`).
    :param: transport: An :class:`RTCDtlsTransport`.
    """
    def __init__(self, trackOrKind, transport):
        if transport.state == 'closed':
            raise InvalidStateError

        if hasattr(trackOrKind, 'kind'):
            self._kind = trackOrKind.kind
            self._track = trackOrKind
        else:
            self._kind = trackOrKind
            self._track = None
        self._ssrc = random32()
        self.__rtp_exited = asyncio.Event()
        self.__rtcp_exited = asyncio.Event()
        self.__started = False
        self.__stopped = asyncio.Event()
        self.__transport = transport

        # stats
        self.__ntp_timestamp = 0
        self.__rtp_timestamp = 0
        self.__octet_count = 0
        self.__packet_count = 0

    @property
    def kind(self):
        return self._kind

    @property
    def track(self):
        """
        The :class:`MediaStreamTrack` which is being handled by the sender.
        """
        return self._track

    @property
    def transport(self):
        """
        The :class:`RTCDtlsTransport` over which media data for the track is
        transmitted.
        """
        return self.__transport

    def replaceTrack(self, track):
        self._track = track

    def setTransport(self, transport):
        self.__transport = transport

    async def send(self, parameters):
        """
        Attempt to set the parameters controlling the sending of media.

        :param: parameters: The :class:`RTCRtpParameters` for the sender.
        """
        if not self.__started:
            asyncio.ensure_future(self._run_rtp(parameters.codecs[0]))
            asyncio.ensure_future(self._run_rtcp())
            self.__started = True

    async def stop(self):
        """
        Irreversibly stop the sender.
        """
        self.__stopped.set()
        if self.__started:
            await asyncio.gather(
                self.__rtp_exited.wait(),
                self.__rtcp_exited.wait())

    async def _run_rtp(self, codec):
        self.__log_debug('- RTP started')
        loop = asyncio.get_event_loop()

        encoder = get_encoder(codec)
        packet = RtpPacket(payload_type=codec.payloadType)
        while not self.__stopped.is_set():
            if self._track:
                frame = await first_completed(self._track.recv(), self.__stopped.wait())
                if frame is True:
                    break
                packet.ssrc = self._ssrc
                payloads = await loop.run_in_executor(None, encoder.encode, frame)
                if not isinstance(payloads, list):
                    payloads = [payloads]
                for i, payload in enumerate(payloads):
                    packet.payload = payload
                    packet.marker = (i == len(payloads) - 1) and 1 or 0
                    try:
                        self.__log_debug('> %s', packet)
                        await self.transport._send_rtp(bytes(packet))
                    except ConnectionError:
                        self.__stopped.set()
                        break
                    self.__ntp_timestamp = datetime_to_ntp(utcnow())
                    self.__rtp_timestamp = packet.timestamp
                    self.__octet_count += len(payload)
                    self.__packet_count += 1
                    packet.sequence_number = seq_plus_one(packet.sequence_number)
                packet.timestamp += encoder.timestamp_increment
            else:
                await asyncio.sleep(0.02)

        self.__log_debug('- RTP finished')
        self.__rtp_exited.set()

    async def _run_rtcp(self):
        self.__log_debug('- RTCP started')

        while not self.__stopped.is_set():
            # The interval between RTCP packets is varied randomly over the
            # range [0.5, 1.5] times the calculated interval.
            sleep = 0.5 + random.random()
            result = await first_completed(asyncio.sleep(sleep), self.__stopped.wait())
            if result is True:
                break

            # RTCP SR
            packets = [RtcpSrPacket(
                ssrc=self._ssrc,
                sender_info=RtcpSenderInfo(
                    ntp_timestamp=self.__ntp_timestamp,
                    rtp_timestamp=self.__rtp_timestamp,
                    packet_count=self.__packet_count,
                    octet_count=self.__octet_count))]

            # RTCP SDES
            if self._cname is not None:
                packets.append(RtcpSdesPacket(chunks=[RtcpSourceInfo(
                    ssrc=self._ssrc,
                    items=[(1, self._cname.encode('utf8'))])]))

            await self._send_rtcp(packets)

        # RTCP BYE
        packet = RtcpByePacket(sources=[self._ssrc])
        await self._send_rtcp([packet])

        self.__log_debug('- RTCP finished')
        self.__rtcp_exited.set()

    async def _send_rtcp(self, packets):
        payload = b''
        for packet in packets:
            self.__log_debug('> %s', packet)
            payload += bytes(packet)

        try:
            await self.transport._send_rtp(payload)
        except ConnectionError:
            pass

    def __log_debug(self, msg, *args):
        logger.debug('sender(%s) ' + msg, self._kind, *args)



import asyncio
import enum
import hmac
import logging
import math
import os
import time
from struct import pack, unpack

import attr
import crcmod.predefined
from pyee import EventEmitter

from .exceptions import InvalidStateError
from .rtcdatachannel import RTCDataChannel, RTCDataChannelParameters
from .utils import first_completed, random32

crc32c = crcmod.predefined.mkPredefinedCrcFun('crc-32c')
logger = logging.getLogger('sctp')

# local constants
COOKIE_LENGTH = 24
COOKIE_LIFETIME = 60
MAX_STREAMS = 65535
USERDATA_MAX_LENGTH = 1200

# protocol constants
SCTP_CAUSE_INVALID_STREAM = 0x0001
SCTP_CAUSE_STALE_COOKIE = 0x0003

SCTP_DATA_LAST_FRAG = 0x01
SCTP_DATA_FIRST_FRAG = 0x02
SCTP_DATA_UNORDERED = 0x04

SCTP_MAX_ASSOCIATION_RETRANS = 10
SCTP_MAX_INIT_RETRANS = 8
SCTP_RTO_INITIAL = 3
SCTP_SEQ_MODULO = 2 ** 16
SCTP_TSN_MODULO = 2 ** 32

RECONFIG_CHUNK = 130
RECONFIG_MAX_STREAMS = 135

# parameters
SCTP_STATE_COOKIE = 0x0007
SCTP_STR_RESET_OUT_REQUEST = 0x000d
SCTP_STR_RESET_RESPONSE = 0x0010
SCTP_STR_RESET_ADD_OUT_STREAMS = 0x0011
SCTP_SUPPORTED_CHUNK_EXT = 0x8008

# data channel constants
DATA_CHANNEL_ACK = 2
DATA_CHANNEL_OPEN = 3

DATA_CHANNEL_RELIABLE = 0

WEBRTC_DCEP = 50
WEBRTC_STRING = 51
WEBRTC_BINARY = 53
WEBRTC_STRING_EMPTY = 56
WEBRTC_BINARY_EMPTY = 57


def chunk_type(chunk):
    return chunk.__class__.__name__


def decode_params(body):
    params = []
    pos = 0
    while pos <= len(body) - 4:
        param_type, param_length = unpack('!HH', body[pos:pos + 4])
        params.append((param_type, body[pos + 4:pos + param_length]))
        pos += param_length + padl(param_length)
    return params


def encode_params(params):
    body = b''
    padding = b''
    for param_type, param_value in params:
        param_length = len(param_value) + 4
        body += padding
        body += pack('!HH', param_type, param_length) + param_value
        padding = b'\x00' * padl(param_length)
    return body


def padl(l):
    return 4 * ((l + 3) // 4) - l


def swapl(i):
    return unpack("<I", pack(">I", i))[0]


def seq_gt(a, b):
    """
    Return True if seq a is greater than b.
    """
    half_mod = (1 << 15)
    return (((a < b) and ((b - a) > half_mod)) or
            ((a > b) and ((a - b) < half_mod)))


def seq_gte(a, b):
    """
    Return True if seq a is greater than or equal to b.
    """
    return (a == b) or seq_gt(a, b)


def seq_plus_one(a):
    return (a + 1) % SCTP_SEQ_MODULO


def tsn_gt(a, b):
    """
    Return True if tsn a is greater than b.
    """
    half_mod = (1 << 31)
    return (((a < b) and ((b - a) > half_mod)) or
            ((a > b) and ((a - b) < half_mod)))


def tsn_gte(a, b):
    """
    Return True if tsn a is greater than or equal to b.
    """
    return (a == b) or tsn_gt(a, b)


def tsn_minus_one(a):
    return (a - 1) % SCTP_TSN_MODULO


def tsn_plus_one(a):
    return (a + 1) % SCTP_TSN_MODULO


class Chunk:
    def __init__(self, flags=0, body=b''):
        self.flags = flags
        self.body = body

    def __bytes__(self):
        body = self.body
        data = pack('!BBH', self.type, self.flags, len(body) + 4) + body
        data += b'\x00' * padl(len(body))
        return data

    def __repr__(self):
        return '%s(flags=%d)' % (chunk_type(self), self.flags)

    @property
    def type(self):
        for k, cls in CHUNK_TYPES.items():
            if isinstance(self, cls):
                return k


class BaseParamsChunk(Chunk):
    def __init__(self, flags=0, body=b''):
        self.flags = flags
        if body:
            self.params = decode_params(body)
        else:
            self.params = []

    @property
    def body(self):
        return encode_params(self.params)


class AbortChunk(BaseParamsChunk):
    pass


class CookieAckChunk(Chunk):
    pass


class CookieEchoChunk(Chunk):
    pass


class DataChunk(Chunk):
    def __init__(self, flags=0, body=b''):
        self.flags = flags
        if body:
            (self.tsn, self.stream_id, self.stream_seq, self.protocol) = unpack('!LHHL', body[0:12])
            self.user_data = body[12:]
        else:
            self.tsn = 0
            self.stream_id = 0
            self.stream_seq = 0
            self.protocol = 0
            self.user_data = b''

    @property
    def body(self):
        body = pack('!LHHL', self.tsn, self.stream_id, self.stream_seq, self.protocol)
        body += self.user_data
        return body

    def __repr__(self):
        return 'DataChunk(flags=%d, tsn=%d, stream_id=%d, stream_seq=%d)' % (
            self.flags, self.tsn, self.stream_id, self.stream_seq)


class ErrorChunk(BaseParamsChunk):
    pass


class HeartbeatChunk(BaseParamsChunk):
    pass


class HeartbeatAckChunk(BaseParamsChunk):
    pass


class BaseInitChunk(Chunk):
    def __init__(self, flags=0, body=b''):
        self.flags = flags
        if body:
            (self.initiate_tag, self.advertised_rwnd, self.outbound_streams,
             self.inbound_streams, self.initial_tsn) = unpack('!LLHHL', body[0:16])
            self.params = decode_params(body[16:])
        else:
            self.initiate_tag = 0
            self.advertised_rwnd = 0
            self.outbound_streams = 0
            self.inbound_streams = 0
            self.initial_tsn = 0
            self.params = []

    @property
    def body(self):
        body = pack(
            '!LLHHL', self.initiate_tag, self.advertised_rwnd, self.outbound_streams,
            self.inbound_streams, self.initial_tsn)
        body += encode_params(self.params)
        return body


class InitChunk(BaseInitChunk):
    pass


class InitAckChunk(BaseInitChunk):
    pass


class ReconfigChunk(BaseParamsChunk):
    pass


class SackChunk(Chunk):
    def __init__(self, flags=0, body=b''):
        self.flags = flags
        self.gaps = []
        self.duplicates = []
        if body:
            self.cumulative_tsn, self.advertised_rwnd, nb_gaps, nb_duplicates = unpack(
                '!LLHH', body[0:12])
            pos = 12
            for i in range(nb_gaps):
                self.gaps.append(unpack('!HH', body[pos:pos + 4]))
                pos += 4
            for i in range(nb_duplicates):
                self.duplicates.append(unpack('!L', body[pos:pos + 4])[0])
                pos += 4
        else:
            self.cumulative_tsn = 0
            self.advertised_rwnd = 0

    @property
    def body(self):
        body = pack('!LLHH', self.cumulative_tsn, self.advertised_rwnd,
                    len(self.gaps), len(self.duplicates))
        for gap in self.gaps:
            body += pack('!HH', *gap)
        for tsn in self.duplicates:
            body += pack('!L', tsn)
        return body

    def __repr__(self):
        return 'SackChunk(flags=%d, advertised_rwnd=%d, cumulative_tsn=%d, gaps=%s)' % (
            self.flags, self.advertised_rwnd, self.cumulative_tsn, self.gaps)


class ShutdownChunk(Chunk):
    def __init__(self, flags=0, body=b''):
        self.flags = flags
        if body:
            self.cumulative_tsn = unpack('!L', body[0:4])[0]
        else:
            self.cumulative_tsn = 0

    @property
    def body(self):
        return pack('!L', self.cumulative_tsn)

    def __repr__(self):
        return 'ShutdownChunk(flags=%d, cumulative_tsn=%d)' % (
            self.flags, self.cumulative_tsn)


class ShutdownAckChunk(Chunk):
    pass


class ShutdownCompleteChunk(Chunk):
    pass


CHUNK_TYPES = {
    0: DataChunk,
    1: InitChunk,
    2: InitAckChunk,
    3: SackChunk,
    4: HeartbeatChunk,
    5: HeartbeatAckChunk,
    6: AbortChunk,
    7: ShutdownChunk,
    8: ShutdownAckChunk,
    9: ErrorChunk,
    10: CookieEchoChunk,
    11: CookieAckChunk,
    14: ShutdownCompleteChunk,
    130: ReconfigChunk,
}


class Packet:
    def __init__(self, source_port, destination_port, verification_tag):
        self.source_port = source_port
        self.destination_port = destination_port
        self.verification_tag = verification_tag
        self.chunks = []

    def __bytes__(self):
        checksum = 0
        data = pack(
            '!HHLL',
            self.source_port,
            self.destination_port,
            self.verification_tag,
            checksum)
        for chunk in self.chunks:
            data += bytes(chunk)

        # calculate checksum
        checksum = swapl(crc32c(data))
        return data[0:8] + pack('!L', checksum) + data[12:]

    @classmethod
    def parse(cls, data):
        if len(data) < 12:
            raise ValueError('SCTP packet length is less than 12 bytes')

        source_port, destination_port, verification_tag, checksum = unpack(
            '!HHLL', data[0:12])

        # verify checksum
        check_data = data[0:8] + b'\x00\x00\x00\x00' + data[12:]
        if checksum != swapl(crc32c(check_data)):
            raise ValueError('SCTP packet has invalid checksum')

        packet = cls(
            source_port=source_port,
            destination_port=destination_port,
            verification_tag=verification_tag)

        pos = 12
        while pos <= len(data) - 4:
            chunk_type, chunk_flags, chunk_length = unpack('!BBH', data[pos:pos + 4])
            chunk_body = data[pos + 4:pos + chunk_length]
            chunk_cls = CHUNK_TYPES.get(chunk_type)
            if chunk_cls:
                packet.chunks.append(chunk_cls(
                    flags=chunk_flags,
                    body=chunk_body))
            pos += chunk_length + padl(chunk_length)
        return packet


# RFC 6525

@attr.s
class StreamResetOutgoingParam:
    request_sequence = attr.ib()
    response_sequence = attr.ib()
    last_tsn = attr.ib()
    streams = attr.ib(default=attr.Factory(list))

    def __bytes__(self):
        data = pack(
            '!LLL',
            self.request_sequence,
            self.response_sequence,
            self.last_tsn)
        for stream in self.streams:
            data += pack('!H', stream)
        return data

    @classmethod
    def parse(cls, data):
        request_sequence, response_sequence, last_tsn = unpack('!LLL', data[0:12])
        streams = []
        for pos in range(12, len(data), 2):
            streams.append(unpack('!H', data[pos:pos + 2])[0])
        return cls(
            request_sequence=request_sequence,
            response_sequence=response_sequence,
            last_tsn=last_tsn,
            streams=streams)


@attr.s
class StreamAddOutgoingParam:
    request_sequence = attr.ib()
    new_streams = attr.ib()

    def __bytes__(self):
        data = pack(
            '!LHH',
            self.request_sequence,
            self.new_streams,
            0)
        return data

    @classmethod
    def parse(cls, data):
        request_sequence, new_streams, reserved = unpack('!LHH', data[0:8])
        return cls(
            request_sequence=request_sequence,
            new_streams=new_streams)


@attr.s
class StreamResetResponseParam:
    response_sequence = attr.ib()
    result = attr.ib()

    def __bytes__(self):
        return pack('!LL', self.response_sequence, self.result)

    @classmethod
    def parse(cls, data):
        response_sequence, result = unpack('!LL', data[0:8])
        return cls(response_sequence=response_sequence, result=result)


RECONFIG_PARAM_TYPES = {
    13: StreamResetOutgoingParam,
    16: StreamResetResponseParam,
    17: StreamAddOutgoingParam
}


class InboundStream:
    def __init__(self):
        self.reassembly = []
        self.sequence_number = 0

    def add_chunk(self, chunk):
        pos = None

        # should never happen, this would mean receiving a chunk
        # for a message that has already been fully re-assembled
        assert seq_gte(chunk.stream_seq, self.sequence_number)

        for i, rchunk in enumerate(self.reassembly):
            # should never happen, the chunk should have been eliminated
            # as a duplicate when _mark_received() is called
            assert rchunk.tsn != chunk.tsn

            if tsn_gt(rchunk.tsn, chunk.tsn):
                pos = i
                break
        if pos is None:
            pos = len(self.reassembly)

        self.reassembly.insert(pos, chunk)

    def pop_messages(self):
        pos = 0
        while pos < len(self.reassembly):
            chunk = self.reassembly[pos]
            if chunk.stream_seq != self.sequence_number:
                break

            if not pos:
                if not (chunk.flags & SCTP_DATA_FIRST_FRAG):
                    break
                expected_tsn = chunk.tsn
                user_data = chunk.user_data
            else:
                if chunk.tsn != expected_tsn:
                    break
                user_data += chunk.user_data

            if (chunk.flags & SCTP_DATA_LAST_FRAG):
                self.reassembly = self.reassembly[pos + 1:]
                self.sequence_number = seq_plus_one(self.sequence_number)
                pos = 0
                yield (chunk.stream_id, chunk.protocol, user_data)
            else:
                pos += 1

            expected_tsn = tsn_plus_one(expected_tsn)


@attr.s
class RTCSctpCapabilities:
    """
    The :class:`RTCSctpCapabilities` dictionary provides information about the
    capabilities of the :class:`RTCSctpTransport`.
    """
    maxMessageSize = attr.ib()
    """
    The maximum size of data that the implementation can send or
    0 if the implementation can handle messages of any size.
    """


class RTCSctpTransport(EventEmitter):
    """
    The :class:`RTCSctpTransport` interface includes information relating to
    Stream Control Transmission Protocol (SCTP) transport.

    :param: transport: An :class:`RTCDtlsTransport`.
    """
    def __init__(self, transport, port=5000):
        if transport.state == 'closed':
            raise InvalidStateError

        super().__init__()
        self.state = self.State.CLOSED
        self.__transport = transport
        self.closed = asyncio.Event()
        self._started = False

        self._loop = asyncio.get_event_loop()
        self._hmac_key = os.urandom(16)

        self.inbound_streams = 0
        self.inbound_streams_max = MAX_STREAMS
        self.outbound_streams = MAX_STREAMS

        self._local_extensions = [RECONFIG_CHUNK]
        self._local_port = port
        self._local_verification_tag = random32()

        self._remote_extensions = []
        self._remote_port = None
        self._remote_verification_tag = 0

        # inbound
        self._advertised_rwnd = 131072
        self._inbound_streams = {}
        self._last_received_tsn = None
        self._sack_duplicates = []
        self._sack_misordered = set()
        self._sack_needed = False

        # outbound
        self._cwnd = 3 * USERDATA_MAX_LENGTH
        self._local_tsn = random32()
        self._last_sacked_tsn = tsn_minus_one(self._local_tsn)
        self._outbound_queue = []
        self._outbound_queue_pos = 0
        self._outbound_stream_seq = {}

        # reconfiguration
        self._reconfig_queue = []
        self._reconfig_request = None
        self._reconfig_request_seq = self._local_tsn
        self._reconfig_response_seq = 0

        # timers
        self._rto = SCTP_RTO_INITIAL
        self._t1_handle = None
        self._t2_handle = None
        self._t3_handle = None

        # data channels
        self._data_channel_id = None
        self._data_channel_queue = []
        self._data_channels = {}

    @property
    def is_server(self):
        return self.transport.transport.role != 'controlling'

    @property
    def port(self):
        """
        The local SCTP port number used for data channels.
        """
        return self._local_port

    @property
    def transport(self):
        """
        The :class:`RTCDtlsTransport` over which SCTP data is transmitted.
        """
        return self.__transport

    @classmethod
    def getCapabilities(cls):
        """
        Retrieve the capabilities of the transport.

        :rtype: RTCSctpCapabilities
        """
        return RTCSctpCapabilities(maxMessageSize=65536)

    def setTransport(self, transport):
        self.__transport = transport

    def start(self, remoteCaps, remotePort):
        """
        Start the transport.
        """
        if not self._started:
            self._remote_port = remotePort
            asyncio.ensure_future(self.__run())
            self._started = True

    async def stop(self):
        """
        Stop the transport.
        """
        if self.state == self.State.CLOSED:
            self.closed.set()
        else:
            await self._abort()

    async def _abort(self):
        """
        Abort the association.
        """
        chunk = AbortChunk()
        try:
            await self._send_chunk(chunk)
        except ConnectionError:
            pass
        self._set_state(self.State.CLOSED)

    async def __run(self):
        """
        The main reception loop.

        It runs until the association reaches the CLOSED state.
        """
        # initialise local channel ID counter
        if self.is_server:
            self._data_channel_id = 0
        else:
            self._data_channel_id = 1

        if not self.is_server:
            chunk = InitChunk()
            chunk.initiate_tag = self._local_verification_tag
            chunk.advertised_rwnd = self._advertised_rwnd
            chunk.outbound_streams = self.outbound_streams
            chunk.inbound_streams = self.inbound_streams_max
            chunk.initial_tsn = self._local_tsn
            self._set_extensions(chunk.params)
            await self._send_chunk(chunk)

            # start T1 timer and enter COOKIE-WAIT state
            self._t1_start(chunk)
            self._set_state(self.State.COOKIE_WAIT)

        while True:
            try:
                data = await first_completed(self.transport.data.recv(), self.closed.wait())
            except ConnectionError:
                self.__log_debug('x Underlying connection closed while reading')
                self._set_state(self.State.CLOSED)
                break
            if data is True:
                break

            try:
                packet = Packet.parse(data)
            except ValueError:
                continue

            # is this an init?
            init_chunk = len([x for x in packet.chunks if isinstance(x, InitChunk)])
            if init_chunk:
                assert len(packet.chunks) == 1
                expected_tag = 0
            else:
                expected_tag = self._local_verification_tag

            # verify tag
            if packet.verification_tag != expected_tag:
                self.__log_debug('Bad verification tag %d vs %d',
                                 packet.verification_tag, expected_tag)
                continue

            # handle chunks
            for chunk in packet.chunks:
                await self._receive_chunk(chunk)

            # send SACK if needed
            if self._sack_needed:
                await self._send_sack()

    def _get_extensions(self, params):
        """
        Gets what extensions are supported by the remote party.
        """
        for k, v in params:
            if k == SCTP_SUPPORTED_CHUNK_EXT:
                self._remote_extensions = list(v)

    def _set_extensions(self, params):
        """
        Sets what extensions are supported by the local party.
        """
        params.append((SCTP_SUPPORTED_CHUNK_EXT, bytes(self._local_extensions)))

    def _get_timestamp(self):
        return int(time.time())

    def _mark_received(self, tsn):
        """
        Mark a data TSN as received.
        """
        # it's a duplicate
        if tsn_gte(self._last_received_tsn, tsn) or tsn in self._sack_misordered:
            self._sack_duplicates.append(tsn)
            return True

        # consolidate misordered entries
        self._sack_misordered.add(tsn)
        for tsn in sorted(self._sack_misordered):
            if tsn == tsn_plus_one(self._last_received_tsn):
                self._last_received_tsn = tsn
            else:
                break

        # filter out obsolete entries
        def is_obsolete(x):
            return tsn_gt(x, self._last_received_tsn)
        self._sack_duplicates = list(filter(is_obsolete, self._sack_duplicates))
        self._sack_misordered = set(filter(is_obsolete, self._sack_misordered))

    async def _receive(self, stream_id, pp_id, data):
        """
        Receive data stream -> ULP.
        """
        await self._data_channel_receive(stream_id, pp_id, data)

    async def _receive_chunk(self, chunk):
        """
        Handle an incoming chunk.
        """
        self.__log_debug('< %s', chunk)

        # server
        if isinstance(chunk, InitChunk) and self.is_server:
            self._last_received_tsn = tsn_minus_one(chunk.initial_tsn)
            self._reconfig_response_seq = tsn_minus_one(chunk.initial_tsn)
            self._remote_verification_tag = chunk.initiate_tag
            self._ssthresh = chunk.advertised_rwnd
            self._get_extensions(chunk.params)

            self.__log_debug('- Peer supports %d outbound streams, %d max inbound streams',
                             chunk.outbound_streams, chunk.inbound_streams)
            self.inbound_streams = min(chunk.outbound_streams, self.inbound_streams_max)
            self.outbound_streams = min(self.outbound_streams, chunk.inbound_streams)

            ack = InitAckChunk()
            ack.initiate_tag = self._local_verification_tag
            ack.advertised_rwnd = self._advertised_rwnd
            ack.outbound_streams = self.outbound_streams
            ack.inbound_streams = self.inbound_streams_max
            ack.initial_tsn = self._local_tsn
            self._set_extensions(ack.params)

            # generate state cookie
            cookie = pack('!L', self._get_timestamp())
            cookie += hmac.new(self._hmac_key, cookie, 'sha1').digest()
            ack.params.append((SCTP_STATE_COOKIE, cookie))
            await self._send_chunk(ack)
        elif isinstance(chunk, CookieEchoChunk) and self.is_server:
            # check state cookie MAC
            cookie = chunk.body
            if (len(cookie) != COOKIE_LENGTH or
               hmac.new(self._hmac_key, cookie[0:4], 'sha1').digest() != cookie[4:]):
                self.__log_debug('x State cookie is invalid')
                return

            # check state cookie lifetime
            now = self._get_timestamp()
            stamp = unpack('!L', cookie[0:4])[0]
            if stamp < now - COOKIE_LIFETIME or stamp > now:
                self.__log_debug('x State cookie has expired')
                error = ErrorChunk()
                error.params.append((SCTP_CAUSE_STALE_COOKIE, b'\x00' * 8))
                await self._send_chunk(error)
                return

            ack = CookieAckChunk()
            await self._send_chunk(ack)
            self._set_state(self.State.ESTABLISHED)

        # client
        if isinstance(chunk, InitAckChunk) and self.state == self.State.COOKIE_WAIT:
            # cancel T1 timer and process chunk
            self._t1_cancel()
            self._last_received_tsn = tsn_minus_one(chunk.initial_tsn)
            self._reconfig_response_seq = tsn_minus_one(chunk.initial_tsn)
            self._remote_verification_tag = chunk.initiate_tag
            self._ssthresh = chunk.advertised_rwnd
            self._get_extensions(chunk.params)

            self.__log_debug('- Peer supports %d outbound streams, %d max inbound streams',
                             chunk.outbound_streams, chunk.inbound_streams)
            self.inbound_streams = min(chunk.outbound_streams, self.inbound_streams_max)
            self.outbound_streams = min(self.outbound_streams, chunk.inbound_streams)

            echo = CookieEchoChunk()
            for k, v in chunk.params:
                if k == SCTP_STATE_COOKIE:
                    echo.body = v
                    break
            await self._send_chunk(echo)

            # start T1 timer and enter COOKIE-ECHOED state
            self._t1_start(echo)
            self._set_state(self.State.COOKIE_ECHOED)
        elif isinstance(chunk, CookieAckChunk) and self.state == self.State.COOKIE_ECHOED:
            # cancel T1 timer and enter ESTABLISHED state
            self._t1_cancel()
            self._set_state(self.State.ESTABLISHED)
        elif (isinstance(chunk, ErrorChunk) and
              self.state in [self.State.COOKIE_WAIT, self.State.COOKIE_ECHOED]):
            self._t1_cancel()
            self._set_state(self.State.CLOSED)
            self.__log_debug('x Could not establish association')
            return

        # common
        elif isinstance(chunk, DataChunk):
            self._sack_needed = True

            # mark as received
            if self._mark_received(chunk.tsn):
                return

            # find stream
            if chunk.stream_id not in self._inbound_streams:
                self._inbound_streams[chunk.stream_id] = InboundStream()
            inbound_stream = self._inbound_streams[chunk.stream_id]

            # defragment data
            inbound_stream.add_chunk(chunk)
            self._advertised_rwnd -= len(chunk.user_data)
            for message in inbound_stream.pop_messages():
                self._advertised_rwnd += len(message[2])
                await self._receive(*message)

        elif isinstance(chunk, SackChunk):
            if tsn_gt(self._last_sacked_tsn, chunk.cumulative_tsn):
                return

            self._last_sacked_tsn = chunk.cumulative_tsn
            done = 0

            # discard acknowledged data
            for i in range(len(self._outbound_queue)):
                schunk = self._outbound_queue[i]
                if tsn_gt(schunk.tsn, self._last_sacked_tsn):
                    break
                done += 1
                self._cwnd = min(self._cwnd + len(schunk.user_data), self._ssthresh)
            if done:
                self._outbound_queue = self._outbound_queue[done:]
                self._outbound_queue_pos = max(0, self._outbound_queue_pos - done)

            if not len(self._outbound_queue):
                # there is no outstanding data, stop T3
                self._t3_cancel()
            elif done:
                # the earliest outstanding chunk was acknowledged, restart T3
                self._t3_handle.cancel()
                self._t3_handle = None
                self._t3_start()

            await self._transmit()
        elif isinstance(chunk, HeartbeatChunk):
            ack = HeartbeatAckChunk()
            ack.params = chunk.params
            await self._send_chunk(ack)
        elif isinstance(chunk, AbortChunk):
            self.__log_debug('x Association was aborted by remote party')
            self._set_state(self.State.CLOSED)
        elif isinstance(chunk, ShutdownChunk):
            self._t2_cancel()
            self._set_state(self.State.SHUTDOWN_RECEIVED)
            ack = ShutdownAckChunk()
            await self._send_chunk(ack)
            self._t2_start(ack)
            self._set_state(self.State.SHUTDOWN_ACK_SENT)
        elif (isinstance(chunk, ShutdownCompleteChunk) and
              self.state == self.State.SHUTDOWN_ACK_SENT):
            self._t2_cancel()
            self._set_state(self.State.CLOSED)
        elif (isinstance(chunk, ReconfigChunk) and self.state == self.State.ESTABLISHED):
            for param in chunk.params:
                cls = RECONFIG_PARAM_TYPES.get(param[0])
                if cls:
                    await self._receive_reconfig_param(cls.parse(param[1]))

    async def _receive_reconfig_param(self, param):
        """
        Handle a RE-CONFIG parameter.
        """
        self.__log_debug('<< %s', param)

        if isinstance(param, StreamResetOutgoingParam):
            # mark closed inbound streams
            for stream_id in param.streams:
                self._inbound_streams.pop(stream_id, None)

                # close data channel
                channel = self._data_channels.get(stream_id)
                if channel:
                    self._data_channel_close(channel)

            # send response
            response_param = StreamResetResponseParam(
                response_sequence=param.request_sequence,
                result=1)
            self._reconfig_response_seq = param.request_sequence

            await self._send_reconfig_param(response_param)
        elif isinstance(param, StreamAddOutgoingParam):
            # increase inbound streams
            self.inbound_streams += param.new_streams

            # send response
            response_param = StreamResetResponseParam(
                response_sequence=param.request_sequence,
                result=1)
            self._reconfig_response_seq = param.request_sequence

            await self._send_reconfig_param(response_param)
        elif isinstance(param, StreamResetResponseParam):
            if (self._reconfig_request and
               param.response_sequence == self._reconfig_request.request_sequence):
                # mark closed streams
                for stream_id in self._reconfig_request.streams:
                    self._outbound_stream_seq.pop(stream_id, None)
                    self._data_channel_closed(stream_id)

                self._reconfig_request = None
                await self._transmit_reconfig()

    async def _send(self, stream_id, pp_id, user_data):
        """
        Send data ULP -> stream.
        """
        stream_seq = self._outbound_stream_seq.get(stream_id, 0)

        fragments = math.ceil(len(user_data) / USERDATA_MAX_LENGTH)
        pos = 0
        for fragment in range(0, fragments):
            chunk = DataChunk()
            chunk.flags = 0
            if fragment == 0:
                chunk.flags |= SCTP_DATA_FIRST_FRAG
            if fragment == fragments - 1:
                chunk.flags |= SCTP_DATA_LAST_FRAG
            chunk.tsn = self._local_tsn
            chunk.stream_id = stream_id
            chunk.stream_seq = stream_seq
            chunk.protocol = pp_id
            chunk.user_data = user_data[pos:pos + USERDATA_MAX_LENGTH]

            pos += USERDATA_MAX_LENGTH
            self._local_tsn = tsn_plus_one(self._local_tsn)
            self._outbound_queue.append(chunk)
        self._outbound_stream_seq[stream_id] = seq_plus_one(stream_seq)

        # transmit outbound data
        await self._transmit()

    async def _send_chunk(self, chunk):
        """
        Transmit a chunk (no bundling for now).
        """
        self.__log_debug('> %s', chunk)
        packet = Packet(
            source_port=self._local_port,
            destination_port=self._remote_port,
            verification_tag=self._remote_verification_tag)
        packet.chunks.append(chunk)
        await self.transport.data.send(bytes(packet))

    async def _send_reconfig_param(self, param):
        chunk = ReconfigChunk()
        for k, cls in RECONFIG_PARAM_TYPES.items():
            if isinstance(param, cls):
                param_type = k
                break
        chunk.params.append((param_type, bytes(param)))

        self.__log_debug('>> %s', param)
        await self._send_chunk(chunk)

    async def _send_sack(self):
        """
        Build and send a selective acknowledgement (SACK) chunk.
        """
        gaps = []
        gap_next = None
        for tsn in sorted(self._sack_misordered):
            pos = (tsn - self._last_received_tsn) % SCTP_TSN_MODULO
            if tsn == gap_next:
                gaps[-1][1] = pos
            else:
                gaps.append([pos, pos])
            gap_next = tsn_plus_one(tsn)

        sack = SackChunk()
        sack.cumulative_tsn = self._last_received_tsn
        sack.advertised_rwnd = self._advertised_rwnd
        sack.duplicates = self._sack_duplicates[:]
        sack.gaps = [tuple(x) for x in gaps]

        await self._send_chunk(sack)

        self._sack_duplicates.clear()
        self._sack_needed = False

    def _set_state(self, state):
        """
        Transition the SCTP association to a new state.
        """
        if state != self.state:
            self.__log_debug('- %s -> %s', self.state, state)
            self.state = state
            if state == self.State.ESTABLISHED:
                asyncio.ensure_future(self._data_channel_flush())
            elif state == self.State.CLOSED:
                self._t1_cancel()
                self._t2_cancel()
                self._t3_cancel()
                self.closed.set()

    # timers

    def _t1_cancel(self):
        if self._t1_handle is not None:
            self.__log_debug('- T1(%s) cancel', chunk_type(self._t1_chunk))
            self._t1_handle.cancel()
            self._t1_handle = None
            self._t1_chunk = None

    def _t1_expired(self):
        self._t1_failures += 1
        self._t1_handle = None
        self.__log_debug('x T1(%s) expired %d', chunk_type(self._t1_chunk), self._t1_failures)
        if self._t1_failures > SCTP_MAX_INIT_RETRANS:
            self._set_state(self.State.CLOSED)
        else:
            asyncio.ensure_future(self._send_chunk(self._t1_chunk))
            self._t1_handle = self._loop.call_later(self._rto, self._t1_expired)

    def _t1_start(self, chunk):
        assert self._t1_handle is None
        self._t1_chunk = chunk
        self._t1_failures = 0
        self.__log_debug('- T1(%s) start', chunk_type(self._t1_chunk))
        self._t1_handle = self._loop.call_later(self._rto, self._t1_expired)

    def _t2_cancel(self):
        if self._t2_handle is not None:
            self.__log_debug('- T2(%s) cancel', chunk_type(self._t2_chunk))
            self._t2_handle.cancel()
            self._t2_handle = None
            self._t2_chunk = None

    def _t2_expired(self):
        self._t2_failures += 1
        self._t2_handle = None
        self.__log_debug('x T2(%s) expired %d', chunk_type(self._t2_chunk), self._t2_failures)
        if self._t2_failures > SCTP_MAX_ASSOCIATION_RETRANS:
            self._set_state(self.State.CLOSED)
        else:
            asyncio.ensure_future(self._send_chunk(self._t2_chunk))
            self._t2_handle = self._loop.call_later(self._rto, self._t2_expired)

    def _t2_start(self, chunk):
        assert self._t2_handle is None
        self._t2_chunk = chunk
        self._t2_failures = 0
        self.__log_debug('- T2(%s) start', chunk_type(self._t2_chunk))
        self._t2_handle = self._loop.call_later(self._rto, self._t2_expired)

    def _t3_expired(self):
        self._t3_handle = None
        self.__log_debug('x T3 expired')

        # retransmit
        self._cwnd = USERDATA_MAX_LENGTH
        self._outbound_queue_pos = 0
        asyncio.ensure_future(self._transmit())

    def _t3_start(self):
        assert self._t3_handle is None
        self.__log_debug('- T3 start')
        self._t3_handle = self._loop.call_later(self._rto, self._t3_expired)

    def _t3_cancel(self):
        if self._t3_handle is not None:
            self.__log_debug('- T3 cancel')
            self._t3_handle.cancel()
            self._t3_handle = None

    async def _transmit(self):
        """
        Transmit outbound data.
        """
        flightsize = 0
        for pos in range(self._outbound_queue_pos):
            flightsize += len(self._outbound_queue[pos].user_data)

        while self._outbound_queue_pos < len(self._outbound_queue):
            chunk = self._outbound_queue[self._outbound_queue_pos]
            flightsize += len(chunk.user_data)
            if flightsize > self._cwnd:
                break
            await self._send_chunk(chunk)
            if not self._t3_handle:
                self._t3_start()
            self._outbound_queue_pos += 1

    async def _transmit_reconfig(self):
        if self._reconfig_queue and not self._reconfig_request:
            streams = self._reconfig_queue[0:RECONFIG_MAX_STREAMS]
            self._reconfig_queue = self._reconfig_queue[RECONFIG_MAX_STREAMS:]
            param = StreamResetOutgoingParam(
                request_sequence=self._reconfig_request_seq,
                response_sequence=self._reconfig_response_seq,
                last_tsn=tsn_minus_one(self._local_tsn),
                streams=streams,
            )
            self._reconfig_request = param
            self._reconfig_request_seq = tsn_plus_one(self._reconfig_request_seq)

            await self._send_reconfig_param(param)

    def _data_channel_close(self, channel, transmit=True):
        """
        Request closing the datachannel by sending an Outgoing Stream Reset Request.
        """
        if channel.readyState not in ['closing', 'closed']:
            channel._setReadyState('closing')
            self._reconfig_queue.append(channel.id)
            if len(self._reconfig_queue) == 1:
                asyncio.ensure_future(self._transmit_reconfig())

    def _data_channel_closed(self, stream_id):
        channel = self._data_channels.pop(stream_id)
        channel._setReadyState('closed')

    async def _data_channel_flush(self):
        """
        Try to flush buffered data to the SCTP layer.

        We wait until the association is established, as we need to know
        whether we are a client or a server to correctly assign an odd/even ID
        to the data channels.
        """
        if self.state != self.State.ESTABLISHED:
            return

        for channel, protocol, user_data in self._data_channel_queue:
            # register channel if necessary
            stream_id = channel.id
            if stream_id is None:
                stream_id = self._data_channel_id
                self._data_channels[stream_id] = channel
                self._data_channel_id += 2
                channel._setId(stream_id)

            # send data
            await self._send(stream_id, protocol, user_data)

        self._data_channel_queue = []

    def _data_channel_open(self, channel):
        data = pack('!BBHLHH', DATA_CHANNEL_OPEN, DATA_CHANNEL_RELIABLE,
                    0, 0, len(channel.label), len(channel.protocol))
        data += channel.label.encode('utf8')
        data += channel.protocol.encode('utf8')
        self._data_channel_queue.append((channel, WEBRTC_DCEP, data))
        asyncio.ensure_future(self._data_channel_flush())

    async def _data_channel_receive(self, stream_id, pp_id, data):
        if pp_id == WEBRTC_DCEP and len(data):
            msg_type = unpack('!B', data[0:1])[0]
            if msg_type == DATA_CHANNEL_OPEN and len(data) >= 12:
                # we should not receive an open for an existing channel
                assert stream_id not in self._data_channels

                # one side should be using even IDs, the other odd IDs
                assert (stream_id % 2) != (self._data_channel_id % 2)

                (msg_type, channel_type, priority, reliability,
                 label_length, protocol_length) = unpack('!BBHLHH', data[0:12])
                pos = 12
                label = data[pos:pos + label_length].decode('utf8')
                pos += label_length
                protocol = data[pos:pos + protocol_length].decode('utf8')

                # register channel
                parameters = RTCDataChannelParameters(label=label, protocol=protocol)
                channel = RTCDataChannel(self, parameters, id=stream_id)
                channel._setReadyState('open')
                self._data_channels[stream_id] = channel

                # send ack
                self._data_channel_queue.append(
                    (channel, WEBRTC_DCEP, pack('!B', DATA_CHANNEL_ACK)))
                await self._data_channel_flush()

                # emit channel
                self.emit('datachannel', channel)
            elif msg_type == DATA_CHANNEL_ACK:
                assert stream_id in self._data_channels
                channel = self._data_channels[stream_id]
                channel._setReadyState('open')
        elif pp_id == WEBRTC_STRING and stream_id in self._data_channels:
            # emit message
            self._data_channels[stream_id].emit('message', data.decode('utf8'))
        elif pp_id == WEBRTC_STRING_EMPTY and stream_id in self._data_channels:
            # emit message
            self._data_channels[stream_id].emit('message', '')
        elif pp_id == WEBRTC_BINARY and stream_id in self._data_channels:
            # emit message
            self._data_channels[stream_id].emit('message', data)
        elif pp_id == WEBRTC_BINARY_EMPTY and stream_id in self._data_channels:
            # emit message
            self._data_channels[stream_id].emit('message', b'')

    async def _data_channel_send(self, channel, data):
        if data == '':
            pp_id, user_data = WEBRTC_STRING_EMPTY, b'\x00'
        elif isinstance(data, str):
            pp_id, user_data = WEBRTC_STRING, data.encode('utf8')
        elif data == b'':
            pp_id, user_data = WEBRTC_BINARY_EMPTY, b'\x00'
        else:
            pp_id, user_data = WEBRTC_BINARY, data

        self._data_channel_queue.append((channel, pp_id, user_data))
        await self._data_channel_flush()

    def __log_debug(self, msg, *args):
        role = self.is_server and 'server' or 'client'
        logger.debug(role + ' ' + msg, *args)

    class State(enum.Enum):
        CLOSED = 1
        COOKIE_WAIT = 2
        COOKIE_ECHOED = 3
        ESTABLISHED = 4
        SHUTDOWN_PENDING = 5
        SHUTDOWN_SENT = 6
        SHUTDOWN_RECEIVED = 7
        SHUTDOWN_ACK_SENT = 8
